@@ -15,8 +15,9 @@ use Unicode::Normalize;
 use Carp;
 
 use Business::CompanyDesignator::Record;
+use Business::CompanyDesignator::SplitResult;
 
-our $VERSION = '0.06';
+our $VERSION = '0.08';
 
 has 'datafile' => ( is => 'ro', default => sub {
   # Development/test version
@@ -28,7 +29,11 @@ has 'datafile' => ( is => 'ro', default => sub {
   return dist_file('Business-CompanyDesignator', 'company_designator.yml');
 });
 
-has [ qw(data assembler regex) ] => ( is => 'ro', lazy_build => 1 );
+# data is the raw dataset as loaded from datafile
+has data => ( is => 'ro', lazy_build => 1 );
+
+# regex_cache is a cache of regexes by language and type, since they're expensive to build
+has 'regex_cache' => ( is => 'ro', isa => 'HashRef', default => sub { {} } );
 
 # abbr_long_map is a hash mapping abbreviations (strings) back to an arrayref of
 # long designators (since abbreviations are not necessarily unique)
@@ -43,12 +48,6 @@ sub _build_data {
   YAML::LoadFile($self->datafile);
 }
 
-sub _build_assembler {
-  my $self = shift;
-  # RA constructor - case insensitive, with match tracking
-  Regexp::Assemble->new->flags('i')->track(1);
-}
-
 sub _build_abbr_long_map {
   my $self = shift;
   my $map = {};
@@ -89,6 +88,7 @@ sub record {
 # Return a list of B::CD::Records for $designator
 sub records {
   my ($self, $designator) = @_;
+  croak "Missing designator" if ! $designator;
   if (exists $self->data->{$designator}) {
     return ( $self->record($designator) );
   }
@@ -102,13 +102,13 @@ sub records {
 
 # Add $string to regex assembler
 sub _add_to_assembler {
-  my ($self, $string, $reference_string) = @_;
+  my ($self, $assembler, $string, $reference_string) = @_;
   $reference_string ||= $string;
 
   # FIXME: RA->add() doesn't work here because of known quantifier-escaping bugs:
   # https://rt.cpan.org/Public/Bug/Display.html?id=50228
   # https://rt.cpan.org/Public/Bug/Display.html?id=74449
-  # $self->assembler->add($string)
+  # $assembler->add($string)
   # Workaround by lexing and using insert()
   my @pattern = map {
     # Periods are treated as optional literals, with optional trailing commas and/or whitespace
@@ -116,47 +116,91 @@ sub _add_to_assembler {
     # Escape other regex metacharacters
     /[()]/ ? "\\$_" : $_
   } split //, $string;
-  $self->assembler->insert(@pattern);
+  $assembler->insert(@pattern);
 
   # Also add pattern => $string mapping to pattern_string_map
   $self->pattern_string_map->{ join '', @pattern } = $reference_string;
 
-  # If $string contains unicode diacritics, also add a version without them for misspellings 
+  # If $string contains unicode diacritics, also add a version without them for misspellings
   if ($string =~ m/\pM/) {
     my $stripped = $string;
     $stripped =~ s/\pM//g;
-    $self->_add_to_assembler($stripped, $string);
+    $self->_add_to_assembler($assembler, $stripped, $string);
   }
 }
 
 # Assemble designator regex
 sub _build_regex {
   my $self = shift;
+  my ($type, $lang) = @_;
+
+  # RA constructor - case insensitive, with match tracking
+  my $assembler = Regexp::Assemble->new->flags('i')->track(1);
 
   while (my ($long, $entry) = each %{ $self->data }) {
+    # If $type is begin, restrict to 'lead' entries
+    next if $type eq 'begin' && ! $entry->{lead};
+    # If $lang is set, restrict to entries that include $lang
+    next if $lang && $entry->{lang} ne $lang;
+
     my $long_nfd = NFD($long);
-    $self->_add_to_assembler($long_nfd);
+    $self->_add_to_assembler($assembler, $long_nfd);
 
     # Add all abbreviations
     if (my $abbr_list = $entry->{abbr}) {
       $abbr_list = [ $abbr_list ] if ! ref $abbr_list;
       for my $abbr (@$abbr_list) {
         my $abbr_nfd = NFD($abbr);
-        $self->_add_to_assembler($abbr_nfd);
+        $self->_add_to_assembler($assembler, $abbr_nfd);
       }
     }
   }
 
-  return $self->assembler->re;
+  return wantarray ? ( $assembler->re, $assembler ) : $assembler->re;
+}
+
+# Regex accessor, returning regexes by type (begin/end) and language (en, es, etc.)
+# $type defaults to 'end', $lang defaults to undef (for all)
+sub regex {
+  my $self = shift;
+  my ($type, $lang) = @_;
+  $type ||= 'end';
+
+  my $cache_key = $type;
+  $cache_key .= "_$lang" if $lang;
+
+  if (my $entry = $self->regex_cache->{ $cache_key }) {
+    return wantarray ? @$entry : $entry->[0];
+  }
+
+  my ($re, $assembler) = $self->_build_regex($type, $lang);
+  $self->regex_cache->{ $cache_key } = [ $re, $assembler ];
+  return wantarray ? ( $re, $assembler ) : $re;
 }
 
 # Helper to return split_designator results
 sub _split_designator_result {
   my $self = shift;
   my ($before, $des, $after, $matched_pattern) = @_;
-  my $matched_string = $self->pattern_string_map->{$matched_pattern}
-    or die "Cannot find matched pattern '$matched_pattern' in pattern_string_map";
-  return map { defined $_ && ! ref $_ ? NFC($_) : $_ } ($before, $des, $after, $matched_string);
+
+  my $des_std;
+  if ($matched_pattern) {
+    $des_std = $self->pattern_string_map->{$matched_pattern}
+      or die "Cannot find matched pattern '$matched_pattern' in pattern_string_map";
+  }
+
+  # Legacy interface - return a simple before / des / after tuple, plus $des_std
+  return map { defined $_ && ! ref $_ ? NFC($_) : '' } ($before, $des, $after, $des_std)
+    if wantarray;
+
+  # New scalar-context interface - return SplitResult object
+  Business::CompanyDesignator::SplitResult->new(
+    before          => NFC($before // ''),
+    designator      => NFC($des // ''),
+    designator_std  => NFC($des_std // ''),
+    after           => NFC($after // ''),
+    records         => [ $des_std ? $self->records(NFC $des_std) : () ],
+  );
 }
 
 # Split $company_name on (the first) company designator, returning a triplet of strings:
@@ -167,22 +211,28 @@ sub _split_designator_result {
 # but "Ltée" for the normalised form
 sub split_designator {
   my $self = shift;
-  my $company_name = shift;
+  my ($company_name, %arg) = @_;
+  my $lang = $arg{lang};
   my $company_name_match = NFD($company_name);
 
-  my $re = $self->regex;
+  my ($re, $assembler) = $self->regex('end', $lang);
+  my ($lead_re, $lead_assembler) = $self->regex('begin', $lang);
 
   # Designators are usually final, so try that first
-  if ($company_name_match =~ m/(.*?)[[:punct:]]*\s+($re)\s*$/) {
-    return $self->_split_designator_result($1, $2, undef, $self->assembler->source($^R));
+  if ($company_name_match =~ m/^\s*(.*?)\p{XPosixPunct}*\s+($re)\s*$/) {
+    return $self->_split_designator_result($1, $2, undef, $assembler->source($^R));
+  }
+  # Not final - check for a lead designator instead (e.g. RU, NL, etc.)
+  elsif ($company_name_match =~ m/^\s*($lead_re)\p{XPosixPunct}*\s*(.*?)\s*$/) {
+    return $self->_split_designator_result(undef, $1, $2, $lead_assembler->source($^R));
   }
   # Not final - check for an embedded designator with trailing content
-  elsif ($company_name_match =~ m/(.*?)[[:punct:]]*\s+($re)(?:\s+(.*?))?$/) {
-    return $self->_split_designator_result($1, $2, $3, $self->assembler->source($^R));
+  elsif ($company_name_match =~ m/(.*?)\p{XPosixPunct}*\s+($re)(?:\s+(.*?))?$/) {
+    return $self->_split_designator_result($1, $2, $3, $assembler->source($^R));
   }
   # No match - return $company_name unchanged
   else {
-    return ($company_name);
+    return $self->_split_designator_result($company_name);
   }
 }
 
@@ -190,6 +240,8 @@ sub split_designator {
 
 __END__
 
+=encoding utf-8
+
 =head1 NAME
 
 Business::CompanyDesignator - module for matching and stripping/manipulating the
@@ -197,7 +249,7 @@ company designators appended to company names
 
 =head1 VERSION
 
-Version: 0.05.
+Version: 0.08.
 
 This module is considered an B<ALPHA> release. Interfaces may change and/or break
 without notice until the module reaches version 1.0.
@@ -206,12 +258,12 @@ without notice until the module reaches version 1.0.
 =head1 SYNOPSIS
 
 Business::CompanyDesignator is a perl module for matching and stripping/manipulating
-the typical company designators appended to company names. It supports both
-long forms (e.g. Corporation, Incorporated, Limited etc.) and abbreviations
-(e.g. Corp., Inc., Ltd., GmbH etc).
+the typical company designators appended (or sometimes, prepended) to company names.
+It supports both long forms (e.g. Corporation, Incorporated, Limited etc.) and
+abbreviations (e.g. Corp., Inc., Ltd., GmbH etc).
 
   use Business::CompanyDesignator;
-  
+
   # Constructor
   $bcd = Business::CompanyDesignator->new;
   # Optionally, you can provide your own company_designator.yml file, instead of the bundled one
@@ -228,14 +280,20 @@ long forms (e.g. Corporation, Incorporated, Limited etc.) and abbreviations
   # Lookup records by abbreviation or long designator (may not be unique)
   @records = $bcd->records($designator);
 
-  # Get a regex for matching designators
+  # Get a regex for matching designators by type ('end'/'begin') and lang
+  # By default, returns 'end' regexes for all languages
   $re = $bcd->regex;
   $company_name =~ $re and say 'designator found!';
   $company_name =~ /$re\s*$/ and say 'final designator found!';
+  my $re_begin_en = $bcd->regex('begin', 'en');
 
   # Split $company_name on designator, returning a ($before, $designator, $after) triplet,
-  # plus the normalised form of the designator matched.
-  ($short_name, $des, $after, $normalised_des) = $bcd->split_designator($company_name);
+  # plus the normalised form of the designator matched (can pass to records(), for example)
+  ($before, $des, $after, $normalised_des) = $bcd->split_designator($company_name);
+
+  # Or in scalar context, return a L<Business::CompanyDesignator::SplitResult> object
+  $res = $bcd->split_designator($company_name, lang => 'en');
+  print join ' / ', $res->designator_std, $res->short_name, $res->extra;
 
 
 =head1 DATASET
@@ -310,28 +368,63 @@ records).
 Use this method for abbreviations, or if you're aren't sure of a
 designator's type.
 
-=head2 regex()
+=head2 regex([$type], [$lang])
 
-Returns a regex that matches all designators from the dataset
-(case-insensitive, non-anchored).
+Returns a regex for all matching designators for $type ('begin'/'end') and
+$lang (iso 639-1 language code e.g. 'en', 'es', de', etc.) from the dataset.
+The regex is case-insensitive and non-anchored.
+
+$type defaults to 'end', so without parameters regex() returns a regex
+matching all designators for all languages.
 
 =head2 split_designator($company_name)
 
 Attempts to split $company_name on (the first) company designator found.
-If found, it returns a list of four items - a triplet of strings from
-$company_name: ( $before, $designator, $after ), plus a normalised version
-of the designator as a fourth element.
 
-  ($short_name, $des, $after_text, $normalised_des) = $bcd->split_designator($company_name);
+In array context split_designator returns a list of four items - a triplet of
+strings from $company_name ( $before, $designator, $after ), plus the
+standardised version of the designator as a fourth element.
+
+  ($short_name, $des, $after_text, $des_std) = $bcd->split_designator($company_name);
+
+In scalar context split_designator returns a L<Business::CompanyDesignator::SplitResult>
+object.
+
+  $res = $bcd->split_designator($company_name);
+
+The initial $des designator (or $res->designator)  is the designator text as
+matched in $company_name, while the final $des_std in array context (or
+$res->designator_std) is the standardised version as found in the dataset.
 
-The initial $des designator is the designator as matched in the text, while
-the second $normalised_des is the normalised version as found in the dataset.
 For instance, "ABC Pty Ltd" would return "Pty Ltd" as the $designator, but
-"Pty. Ltd." as the normalised form, and the latter would be what you
+"Pty. Ltd." as the stardardised form, and the latter would be what you
 would find in designators() or would lookup with records(). Similarly,
-"Accessoires XYZ Ltee" (misspelt without the grave accent) would still be
-matched, returning "Ltee" (as found) for the $designator, but "Ltée" as the
-normalised form.
+"Accessoires XYZ Ltee" (without the french acute) would match, returning
+"Ltee" (as found) for the $designator, but "Ltée" as the standardised form.
+
+split_designator also accepts an optional 'lang' parameter, which would be
+the ISO 639-1 language code for $company_name. If this is given,
+split_designator will only match designators for that language, which can
+improve the accuracy of the split.
+
+Note that split_designator won't always get the split right. It checks for
+final designators first, then leading ones, and then finally looks for embedded
+designators. This allows names like these to picked up:
+
+    Amerihealth Insurance Company of NJ
+    Trenkwalder Personal AG Schweiz
+    Vicente Campano S L (COMERCIAL VICAM)
+    Gvozdika, gostinitsa OOO ""Eko-Treyd""
+
+but it can also detect designators that are false positives e.g.
+
+    Dr S L Ledingham - Beaumont Street Practice
+
+One way to reduce this is to specify the optional 'lang' parameter if you know
+the language your company names are in. This should reduce the number of false
+positives by making fewer designators available to match on, but it doesn't
+eliminate the issue altogether.
+
 
 =head1 SEE ALSO
 
@@ -343,7 +436,7 @@ Gavin Carr <gavin@profound.net>
 
 =head1 COPYRIGHT AND LICENCE
 
-Copyright (C) 2013 Gavin Carr and Profound Networks.
+Copyright (C) 2013-2015 Gavin Carr
 
 This library is free software; you can redistribute it and/or modify it
 under the same terms as Perl itself.
